VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "cHashD"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit 'Olaf Schmidt in August 2016
 
Private Type SAFEARRAY1D
  cDims As Integer
  fFeatures As Integer
  cbElements As Long
  cLocks As Long
  pvData As Long
  cElements1D As Long
  lLbound1D As Long
End Type
Private Declare Sub BindArray Lib "kernel32" Alias "RtlMoveMemory" (PArr() As Any, pSrc&, Optional ByVal CB& = 4)
Private Declare Sub ReleaseArray Lib "kernel32" Alias "RtlMoveMemory" (PArr() As Any, Optional pSrc& = 0, Optional ByVal CB& = 4)

Private Declare Function VariantCopy Lib "oleaut32" (Dst As Any, Src As Any) As Long
Private Declare Function VariantCopyInd Lib "oleaut32" (Dst As Any, Src As Any) As Long
 
Private Type HashTableEntry
  Count As Long
  DataIndexes() As Long
End Type

Public Enum eKeyMode
  eUseStrings
  eUseIntegers
  eUseDoubles
End Enum

Private W() As Integer, saW As SAFEARRAY1D

Private mCompareMode As VbCompareMethod, mUniqueKeys As Boolean, mKeyMode As eKeyMode
Private mCount As Long, mDTUB As Long, mHashTableSize As Long, mLastExpectedCount As Long
Private mLastKey, mLastIdx As Long
Private mKeys() As String, mCurs() As Currency, mDbls() As Double
Private HashTable() As HashTableEntry, mValues() As Variant

Private Sub Class_Initialize()
  saW.cDims = 1
  saW.cbElements = 2
  BindArray W, VarPtr(saW)
 
  mUniqueKeys = True
  mCompareMode = vbBinaryCompare
  mLastExpectedCount = 32768 'set the expected default HashTable-Size
  ReInit
End Sub
Private Sub Class_Terminate()
  ReleaseArray W
End Sub

Public Sub ReInit(Optional ByVal ExpectedCount As Long)
  If ExpectedCount <= 0 Then ExpectedCount = mLastExpectedCount
  If ExpectedCount < 100 Then ExpectedCount = 100
  mLastExpectedCount = ExpectedCount
  
  mHashTableSize = 8
  Do Until mHashTableSize * 4 > ExpectedCount: mHashTableSize = mHashTableSize * 2: Loop
  ReDim HashTable(0 To mHashTableSize - 1)
  
  Dim i As Long
  For i = 0 To UBound(HashTable): ReDim HashTable(i).DataIndexes(0 To 4): Next
  
  Erase mKeys: Erase mCurs: Erase mDbls
  Select Case mKeyMode
    Case eUseStrings:  mDTUB = 16: ReDim mKeys(0 To mDTUB)
    Case eUseIntegers: mDTUB = 16: ReDim mCurs(0 To mDTUB)
    Case eUseDoubles:  mDTUB = 16: ReDim mDbls(0 To mDTUB)
  End Select
  ReDim mValues(0 To mDTUB)
  
  mCount = 0
  mLastIdx = 0
End Sub

Public Sub RemoveAll()
  ReInit
End Sub

Public Property Get Count() As Long
  Count = mCount
End Property

Public Property Get UniqueKeys() As Boolean
  UniqueKeys = mUniqueKeys
End Property
Public Property Let UniqueKeys(ByVal RHS As Boolean)
  mUniqueKeys = RHS
End Property

Public Property Get CompareMode() As VbCompareMethod
  CompareMode = mCompareMode
End Property
Public Property Let CompareMode(ByVal RHS As VbCompareMethod)
  mCompareMode = RHS
End Property

Public Property Get KeyMode() As eKeyMode
  KeyMode = mKeyMode
End Property
Public Property Let KeyMode(ByVal RHS As eKeyMode)
  mKeyMode = RHS
  ReInit
End Property

Public Function Keys()
  If mCount Then mDTUB = mCount - 1 Else Keys = Split(vbNullString): Exit Function
  
  Select Case mKeyMode
    Case eUseStrings:  ReDim Preserve mKeys(0 To mDTUB): Keys = mKeys
    Case eUseIntegers: ReDim Preserve mCurs(0 To mDTUB): Keys = mCurs
    Case eUseDoubles:  ReDim Preserve mDbls(0 To mDTUB): Keys = mDbls
  End Select
End Function
Public Function Items()
  If mCount = 0 Then Items = Array(): Exit Function
  
  ReDim Preserve mValues(0 To mCount - 1): mDTUB = mCount - 1
  Items = mValues
End Function

Public Function Exists(Key) As Boolean
  mLastIdx = FindIndex(Key, HashTable(CalcHash(Key))) + 1
  If mLastIdx Then mLastKey = Key
  Exists = mLastIdx
End Function
Public Function IndexByKey(Key) As Long
  IndexByKey = FindIndex(Key, HashTable(CalcHash(Key)))
End Function

Public Sub Add(Key, Item)
Dim HashValue As Long, UB As Long
    HashValue = CalcHash(Key)
  If mUniqueKeys Then If FindIndex(Key, HashTable(HashValue)) >= 0 Then Err.Raise 457
  
  'add the new Pair, prolonging the Keys- and Values-arrays
  If mDTUB < mCount Then
     mDTUB = (mDTUB + 16) * 1.6
     Select Case mKeyMode
       Case eUseStrings:  ReDim Preserve mKeys(0 To mDTUB)
       Case eUseIntegers: ReDim Preserve mCurs(0 To mDTUB)
       Case eUseDoubles:  ReDim Preserve mDbls(0 To mDTUB)
     End Select
     ReDim Preserve mValues(0 To mDTUB)
  End If
  Select Case mKeyMode
    Case eUseStrings:  mKeys(mCount) = Key
    Case eUseIntegers: mCurs(mCount) = Key
    Case eUseDoubles:  mDbls(mCount) = Key
  End Select
  
  VariantCopyInd ByVal VarPtr(mValues(mCount)), ByVal VarPtr(Item)
  
  'add the new DataIndex to the proper Hash-Buckets .DataIndexes-Array
  With HashTable(HashValue)
    UB = UBound(.DataIndexes)
    If UB < .Count Then UB = UB + UB: ReDim Preserve .DataIndexes(0 To UB)
    .DataIndexes(.Count) = mCount
    .Count = .Count + 1
  End With
  
  mCount = mCount + 1
End Sub

Public Property Get KeyByIndex(ByVal IndexZeroBased As Long)
  If IndexZeroBased < 0 Or IndexZeroBased >= mCount Then Err.Raise 9
  Select Case mKeyMode
    Case eUseStrings:  KeyByIndex = mKeys(IndexZeroBased)
    Case eUseIntegers: KeyByIndex = mCurs(IndexZeroBased)
    Case eUseDoubles:  KeyByIndex = mDbls(IndexZeroBased)
  End Select
End Property

Public Property Get ItemByIndex(ByVal IndexZeroBased As Long)
  If IndexZeroBased < 0 Or IndexZeroBased >= mCount Then Err.Raise 9
  VariantCopy ItemByIndex, ByVal VarPtr(mValues(IndexZeroBased))
End Property
Public Property Let ItemByIndex(ByVal IndexZeroBased As Long, RHS)
  If IndexZeroBased < 0 Or IndexZeroBased >= mCount Then Err.Raise 9
  VariantCopyInd ByVal VarPtr(mValues(IndexZeroBased)), ByVal VarPtr(RHS)
End Property
Public Property Set ItemByIndex(ByVal IndexZeroBased As Long, RHS)
  ItemByIndex(IndexZeroBased) = RHS
End Property

Public Property Get Item(Key)
Attribute Item.VB_UserMemId = 0
Dim Index As Long
  If mLastIdx Then
    If mLastKey = Key Then VariantCopy Item, ByVal VarPtr(mValues(mLastIdx - 1)): Exit Property
  End If
  Index = FindIndex(Key, HashTable(CalcHash(Key)))
  If Index >= 0 Then VariantCopy Item, ByVal VarPtr(mValues(Index))
End Property
Public Property Let Item(Key, RHS)
Dim Index As Long
  Index = FindIndex(Key, HashTable(CalcHash(Key)))
  If Index = -1 Then Add Key, RHS Else VariantCopyInd ByVal VarPtr(mValues(Index)), ByVal VarPtr(RHS)
End Property
Public Property Set Item(Key, RHS)
  Item(Key) = RHS
End Property

Private Function CalcHash(Key) As Long
Dim l As Long, i As Long, S As String, C As Currency, D As Double
  If mKeyMode = eUseStrings Then
    If mCompareMode Then
      S = LCase$(Key)
      saW.cElements1D = Len(S)
      saW.pvData = StrPtr(S)
    ElseIf VarType(Key) = vbString Then
      saW.cElements1D = Len(Key)
      saW.pvData = StrPtr(Key)
    Else
      S = Key
      saW.cElements1D = Len(S)
      saW.pvData = StrPtr(S)
    End If
    If saW.cElements1D = 0 Then Exit Function
  ElseIf mKeyMode = eUseIntegers Then
    C = Key: saW.cElements1D = 4: saW.pvData = VarPtr(C)
  Else
    D = Key: saW.cElements1D = 4: saW.pvData = VarPtr(D)
  End If
 
  l = 65233
    For i = saW.cElements1D - 1 To 0 Step -1
      l = (l + W(i) + i) * 3727 And &HFFFF&   '
    Next
  CalcHash = (l + W(0) + saW.cElements1D) And (mHashTableSize - 1)
End Function

Private Function FindIndex(Key, H As HashTableEntry) As Long  'return -1, when no Key can be found
  Dim i As Long
  Select Case mKeyMode
    Case eUseStrings
      For i = 0 To H.Count - 1
        If StrComp(Key, mKeys(H.DataIndexes(i)), mCompareMode) = 0 Then
          FindIndex = H.DataIndexes(i): Exit Function
        End If
      Next
    Case eUseIntegers
      For i = 0 To H.Count - 1
        If Key = mCurs(H.DataIndexes(i)) Then FindIndex = H.DataIndexes(i): Exit Function
      Next
    Case eUseDoubles
      For i = 0 To H.Count - 1
        If Key = mDbls(H.DataIndexes(i)) Then FindIndex = H.DataIndexes(i): Exit Function
      Next
  End Select
  FindIndex = -1
End Function
 
Friend Sub CheckHashDistribution()
Dim i As Long, Count As Long, cc As Long, Min As Long, Max As Long
  Min = &H7FFFFFFF
  For i = 0 To UBound(HashTable)
    Count = HashTable(i).Count
    If Count Then
      If Min > Count Then Min = Count
      If Max < Count Then Max = Count
      cc = cc + 1
    End If
  Next
  Debug.Print "Distribution over a HashTable with"; UBound(HashTable) + 1; "slots:"
  Debug.Print "Used-HashSlots:"; cc
  Debug.Print "Min-Entries:"; Min
  Debug.Print "Max-Entries:"; Max
End Sub

